/*
 * game.h
 *
 * Author: Peter Sutton
 *
 * The game has 16 rows, numbered 0 to 15 from the bottom. 
 * The car is always in rows 1 and 2 (two pixels in size).
 * The background scrolls down. There are 8 columns, numbered
 * 0 to 7 from the left.
 *
 * The functions in this module will update the LED matrix
 * display as required. Note that the LED matrix display has
 * a different understanding of rows and columns. See the comment
 * at the top of game.c for details.
 */ 

#ifndef GAME_H_
#define GAME_H_

#include <stdint.h>

// Reset the game. Get the background ready and place the car in the 
// initial position. Output this to the LED matrix.
void init_game(void);
void set_alive(void);

// Put a car at the starting position. (This would typically be called after
// a restart.)
void put_car_at_start(void);

/////////////////////////////////// MOVE FUNCTIONS /////////////////////////
// has_car_crashed() should be checked after calling one of these to see
// if the move succeeded or not

// Move the car one column left
// This function must NOT be called if the car is in column 0.
// Failure may occur if the car collides with the background.
void move_car_left(void);

// Move the car one column right
// This function must NOT be called if the car is in column 7.
// Failure may occur if the car collides with the background.
void move_car_right(void);

/////////////////////// CAR / GAME STATUS ///////////////////////////////////
// Get the column that the car is currently in (will return a number
// between 0 and 7 inclusive)
uint8_t get_car_column(void);

// Return true if the car has crashed, false otherwise
uint8_t has_car_crashed(void);

// Return true if the car has finished the lap, false otherwise
uint8_t has_lap_finished(void);

/////////////////////// LIVES ////////////////////////////////////////////////

void init_lives(void);

uint8_t get_lives(void);

void player_killed(void);

void bonus_life(void);

/////////////////////// LAPS /////////////////////////////////////////////////

void init_laps(void);

void inc_laps(void);

uint8_t get_laps(void);

/////////////////////// SPEED ////////////////////////////////////////////////

void set_speed(uint8_t);

void inc_speed(void);

void dec_speed(void);

uint8_t get_speed(void);

uint32_t new_scroll_speed(void);

/////////////////////// UPDATE FUNCTIONS /////////////////////////////////////
// Scroll the background by one row and update the display. Note that this
// may cause the car to crash.
void scroll_background(void);

////////////////////// POWERUP INTERFACE //////////////////////////////////////
void place_powerup(uint32_t);

void flash_powerup(void);

////////////////////// SAVE/LOAD//////////////////////////////////////////////

void set_data(void);

uint8_t get_invincible(void);
uint32_t get_invinc_timer(void);
uint8_t get_puprow(void);
uint8_t get_pupcol(void);
uint8_t get_scroll_position(void);
void save_temp_data(void);
void load_temp_data(void);
void set_lives(uint8_t);
void set_speed(uint8_t);
void set_laps(uint8_t);
void set_invincible(uint8_t);
void set_scroll_position(uint8_t);
void set_car_column(uint8_t);
void set_puprow(uint8_t);
void set_pupcol(uint8_t);

void set_invinc_timer(uint32_t);

void redraw_background(uint8_t);


////////////////////////////////////////////////////////////////////////////////
void shoot(void);
void move_bullet(uint8_t);

#endif /* GAME_H_ */