/*
 * ledmatrix.c
 *
 * Author: Peter Sutton
 * 
 * See the LED matrix Reference for details of the SPI commands used.
 */ 

#include <avr/io.h>
#include "ledmatrix.h"
#include "spi.h"
#include "serialio.h"
#include "terminalio.h"
#include <avr/pgmspace.h>
#include <stdio.h>


#define CMD_UPDATE_ALL 0x00
#define CMD_UPDATE_PIXEL 0x01
#define CMD_UPDATE_ROW 0x02
#define CMD_UPDATE_COL 0x03
#define CMD_SHIFT_DISPLAY 0x04
#define CMD_CLEAR_SCREEN 0x0F

void setup_terminal_display(void);
void display_update_column(uint8_t, MatrixColumn);
void display_update_pixel(uint8_t, uint8_t, PixelColour);
void display_shift(void);
char coords[8][16];
uint8_t active = 0;

void ledmatrix_setup(void) {
	// Setup SPI - we divide the clock by 128.
	// (This speed guarantees the SPI buffer will never overflow.)
	spi_setup_master(128);
}

void ledmatrix_update_all(MatrixData data) {
	(void)spi_send_byte(CMD_UPDATE_ALL);
	for(uint8_t y=0; y<MATRIX_NUM_ROWS; y++) {
		for(uint8_t x=0; x<MATRIX_NUM_COLUMNS; x++) {
			(void)spi_send_byte(data[x][y]);
		}
	}
}

void ledmatrix_update_pixel(uint8_t x, uint8_t y, PixelColour pixel, uint8_t terminal) {
	(void)spi_send_byte(CMD_UPDATE_PIXEL);
	(void)spi_send_byte( ((y & 0x07)<<4) | (x & 0x0F));
	(void)spi_send_byte(pixel);
	if (active) {
		display_update_pixel(x, y, pixel);
	}
}

void ledmatrix_update_row(uint8_t y, MatrixRow row) {
	(void)spi_send_byte(CMD_UPDATE_ROW);
	(void)spi_send_byte(y & 0x07);	// row number
	for(uint8_t x = 0; x<MATRIX_NUM_COLUMNS; x++) {
		(void)spi_send_byte(row[x]);
	}
}

void ledmatrix_update_column(uint8_t x, MatrixColumn col, uint8_t terminal) {
	(void)spi_send_byte(CMD_UPDATE_COL);
	(void)spi_send_byte(x & 0x0F); // column number
	for(uint8_t y = 0; y<MATRIX_NUM_ROWS; y++) {
		(void)spi_send_byte(col[y]);
	}
	if (terminal && active) {
		display_update_column(x, col);
	}
}

void ledmatrix_shift_display_left(void) {
	(void)spi_send_byte(CMD_SHIFT_DISPLAY);
	(void)spi_send_byte(0x02);
}

void ledmatrix_shift_display_right(void) {
	(void)spi_send_byte(CMD_SHIFT_DISPLAY);
	(void)spi_send_byte(0x01);
	if (active) {
		display_shift();
	}
}

void ledmatrix_shift_display_up(void) {
	(void)spi_send_byte(CMD_SHIFT_DISPLAY);
	(void)spi_send_byte(0x08);
}

void ledmatrix_shift_display_down(void) {
	(void)spi_send_byte(CMD_SHIFT_DISPLAY);
	(void)spi_send_byte(0x04);
}

void ledmatrix_clear(void) {
	(void)spi_send_byte(CMD_CLEAR_SCREEN);
	setup_terminal_display();
}

void setup_terminal_display(void) {
	for (uint8_t x = 0; x < 16; x++) {
		for (uint8_t y = 0; y < 8; y++) {
			coords[y][x] = 'x';
		}
	}
	set_scroll_region(10,26);
}

void display_default(void) {
	for (uint8_t x = 0; x < 16; x++) {
		move_cursor(53, 10+x);
		for (uint8_t y = 0; y < 8; y++) {
			printf_P(PSTR("%c"), coords[y][x]);
		}
	}
}



void display_update_column(uint8_t x, MatrixColumn col) {
	for (uint8_t y = 0; y < 8; y++) {
		if (col[y]) {
			set_display_attribute(FG_GREEN);
		} else {
			set_display_attribute(FG_BLACK);
		}
		move_cursor(53 + y, 10+x);
		printf_P(PSTR("%c"), coords[y][x]);
	}
	set_display_attribute(FG_WHITE);
}

void display_update_pixel(uint8_t x, uint8_t y, PixelColour pixel) {
	if (pixel == COLOUR_YELLOW) {
		set_display_attribute(FG_YELLOW);	
		move_cursor(53 + y, 10 + x );
	} else if (pixel == COLOUR_LIGHT_ORANGE) {
		set_display_attribute(FG_BLUE);
	} else if (pixel == COLOUR_RED) {
		set_display_attribute(FG_RED);
	}
	move_cursor(53 + y, 10 + x);
	printf_P(PSTR("%c"), coords[y][x]);
	set_display_attribute(FG_WHITE);
}

void display_shift(void) {
	move_cursor(53,10);
	scroll_down();
}

void set_active(uint8_t setting) {
	active = setting;
}