/*
 * RallyProject.c
 *
 * Main file
 *
 * Author: Peter Sutton. Modified by <YOUR NAME HERE>
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>		// For random()
#include <avr/eeprom.h>

#include "buzzer.h"
#include "ledmatrix.h"
#include "scrolling_char_display.h"
#include "buttons.h"
#include "serialio.h"
#include "terminalio.h"
#include "score.h"
#include "timer0.h"
#include "game.h"
#include "display.h"

#define F_CPU 8000000L
#include <util/delay.h>
#include <string.h>

// Function prototypes - these are defined below (after main()) in the order
// given here
// INIT HARDWARE
void initialise_hardware(void);
uint8_t splash_screen(void);
// NEW GAME HANDLING
void new_game(void);
void start_over(void);
// MAIN LOOP RUNNING WHILE GAME PLAYS
uint8_t play_game(void);
// IF USER PRESSES N KEY PLAY GAME RETURNS 1 FOR USER INTERRUPT
void user_interrupt(uint8_t);

// GAME END HANDLING
uint8_t handle_game_over(uint8_t);
uint8_t handle_new_lap(void);
// SAVE GAME HANDLING
void save_game(void);
void open_game(void);

// ASCII code for Escape character
#define ESCAPE_CHAR 27
uint8_t serial_input;
uint32_t timer;
uint8_t paused;
uint32_t current_time;
uint32_t last_move_time;

// Strings for use in scrolling display
char* title = "RALLYRACER40817754";
char level[10];
/////////////////////////////// main //////////////////////////////////
int main(void) {
	uint8_t usr;
	// Setup hardware and call backs. This will turn on 
	// interrupts.
	initialise_hardware();
	
	// Show the splash screen message. Returns when display
	// is complete
	uint8_t game_loaded;
	game_loaded = splash_screen();
	set_active(1);
	while(1) {
		usr = 0;
		if (!game_loaded) {
			new_game();
		}
		usr = play_game();
		if (usr == 1) {
			user_interrupt(usr);
		}
		if (usr != 1) {
			game_loaded = handle_game_over(usr);
		}
	}
}
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// HARDWARE INIT, LED MATRIX MESSAGE HANDLING
/////////////////////////////////////////////////////////////////////
void initialise_hardware(void) {
	ledmatrix_setup();
	init_button_interrupts();
	allocate_memory();
	// Setup serial port for 19200 baud communication with no echo
	// of incoming characters
	init_serial_stdio(19200,0);
	
	init_timer0();
	
	// Turn on global interrupts
	sei();
	
	// Set the default highscore (from eeprom)
	set_highscores();
}

uint8_t splash_screen(void) {
	// Reset display attributes and clear terminal screen then output a message
	set_display_attribute(TERM_RESET);
	clear_terminal();
	hide_cursor();	// We don't need to see the cursor when we're just doing output
	move_cursor(3,3);
	printf_P(PSTR("RallyRacer"));
	
	move_cursor(3,5);
	set_display_attribute(FG_GREEN);	// Make the text green
	printf_P(PSTR("CSSE2010/7201 project by Joseph Ray - 40817754"));	
	set_display_attribute(FG_WHITE);	// Return to default colour (White)
	move_cursor(3,7);
	printf_P(PSTR("Press 'n' for new game"));
	move_cursor(3,20);
	printf_P(PSTR("To see the scrolling terminal display, change line 73 (in the main function)"));
	move_cursor(3,21);
	printf_P(PSTR("of project.c to 'set_active(1);'"));
	display_highscore();

	// Output the scrolling message to the LED matrix
	// and wait for a push button to be pushed.
	ledmatrix_clear();
	
	// Orange message the first time through
	set_text_colour(COLOUR_ORANGE);
	
	while(1) {
		set_scrolling_display_text(title);
		// Scroll the message until it has scrolled off the 
		// display or a button is pushed. We pause for 130ms between each scroll.
		while(scroll_display()) {
			
			if (serial_input_available()) {
				serial_input = fgetc(stdin);
			}
			_delay_ms(130);
			
			//check for game start, button pushed or ['n', 'N'] typed
			if((button_pushed() != -1) || serial_input == 'n' || serial_input == 'N') {
				return 0;
			} else if(serial_input == 'o' || serial_input == 'O') {
				stop_timer0();
				open_game();
				display_ingame_status();
				redraw_background(1);
				flash_powerup();
				_delay_ms(500);
				start_timer0();
				start_timer1();
				return 1;
			} 
		}
		// Message has scrolled off the display. Change colour
		// to a random colour and scroll again.
		switch(random()%4) {
			case 0: set_text_colour(COLOUR_LIGHT_ORANGE); break;
			case 1: set_text_colour(COLOUR_RED); break;
			case 2: set_text_colour(COLOUR_YELLOW); break;
			case 3: set_text_colour(COLOUR_GREEN); break;
		}
	}
	
}
////////////////////////////////////////////////////////////////
// NEW GAME/LAP HANDLING
////////////////////////////////////////////////////////////////

void new_game(void) {
	// Clear the serial terminal
	display_ingame_status();
	setup_terminal_display();
	
	uint32_t seed;
	seed = get_clock_ticks();
	place_powerup(seed);
	// Initialise the game and display
	init_laps();
	init_lives();
	set_speed(5);
	init_game();
	
	
	
	// Initialise the score
	init_score();
	
	// Clear a button push or serial input if any are waiting
	// (The cast to void means the return value is ignored.)
	(void)button_pushed();
	clear_serial_input_buffer();
	
	// Delay for half a second
	reset_count();
	update_score();
	_delay_ms(500);
	
}

void start_over(void) {
	// Clear the serial terminal
	display_ingame_status();
	setup_terminal_display();
	
	uint32_t seed;
	seed = get_clock_ticks();
	place_powerup(seed);
	// Initialise the game and display
	init_game();
	set_speed(5);

	
	
	// Clear a button push or serial input if any are waiting
	// (The cast to void means the return value is ignored.)
	(void)button_pushed();
	clear_serial_input_buffer();
	
	// Delay for half a second
	reset_count();
	_delay_ms(500);
}

/////////////////////////////////////////////////////////////////
// WHILE GAME IS RUNNING
////////////////////////////////////////////////////////////////

uint8_t play_game(void) {
	uint32_t scroll_speed, flash_timer;
	int8_t button;
	uint8_t scrolled=0;
	char escape_sequence_char;
	uint8_t characters_into_escape_sequence = 0;
	uint32_t bangtime;
	
	// Get the current time and remember this as the last time the background scrolled.
	current_time = get_clock_ticks();
	last_move_time = current_time;
	flash_timer = current_time;
	bangtime = current_time;
	init_timer1();
	scroll_speed = new_scroll_speed();
	// We play the game while the car hasn't crashed
	while(!has_car_crashed()) {
		update_timer();
		// check if we need to flash the powerup
		
		//check_LCD_display();
		//throw_LCD_display();
		
		if (current_time > 50 + flash_timer) {
			move_bullet(scrolled);
			flash_powerup();
			flash_timer = current_time;
			scrolled = 0;
		}
		
		if (current_time > 300 + bangtime) {	
			// clear "BANG!!" from the screen
			move_cursor(15,9);
			printf_P(PSTR("      "));
		}
		// Check for input - which could be a button push or serial input.
		// Serial input may be part of an escape sequence, e.g. ESC [ D
		// is a left cursor key press. We will be processing each character
		// independently and can't do anything until we get the third character.
		// At most one of the following three variables will be set to a value 
		// other than -1 if input is available.
		// (We don't initalise button to -1 since button_pushed() will return -1
		// if no button pushes are waiting to be returned.)
		// Button pushes take priority over serial input. If there are both then
		// we'll retrieve the serial input the next time through this loop
		serial_input = -1;
		escape_sequence_char = -1;
		button = button_pushed();
		
		if(button == -1) {
			// No push button was pushed, see if there is any serial input
			if(serial_input_available()) {
				// Serial data was available - read the data from standard input
				serial_input = fgetc(stdin);
				// Check if the character is part of an escape sequence
				if(characters_into_escape_sequence == 0 && serial_input == ESCAPE_CHAR) {
					// We've hit the first character in an escape sequence (escape)
					characters_into_escape_sequence++;
					serial_input = -1; // Don't further process this character
				} else if(characters_into_escape_sequence == 1 && serial_input == '[') {
					// We've hit the second character in an escape sequence
					characters_into_escape_sequence++;
					serial_input = -1; // Don't further process this character
				} else if(characters_into_escape_sequence == 2) {
					// Third (and last) character in the escape sequence
					escape_sequence_char = serial_input;
					serial_input = -1;  // Don't further process this character - we
										// deal with it as part of the escape sequence
					characters_into_escape_sequence = 0;
				} else {
					// Character was not part of an escape sequence (or we received
					// an invalid second character in the sequence). We'll process 
					// the data in the serial_input variable.
					characters_into_escape_sequence = 0;
				}
			}
		}
		
		// Process the input. 
		if(button==3 || escape_sequence_char=='D' || serial_input=='L' || serial_input=='l') {
			// Attempt to move left
			count_moves();
			move_car_left();
		} else if(button==0 || escape_sequence_char=='C' || serial_input=='R' || serial_input=='r') {
			// Attempt to move right
			count_moves();
			move_car_right();
		} else if(button==2 || escape_sequence_char=='A' || serial_input=='U' || serial_input=='u') {
			// speeds us up
			inc_speed();
			update_speed();
			scroll_speed = new_scroll_speed();
		} else if(button==1 || escape_sequence_char=='B' || serial_input=='D' || serial_input=='d') {
			// slows us down
			dec_speed();
			update_speed();
			scroll_speed = new_scroll_speed();
		} else if (serial_input == ' ') {
			move_cursor(15,9);
			printf_P(PSTR("BANG!!"));
			bangtime = current_time;
			shoot();
		} else if(serial_input == 's' || serial_input == 'S') {
			save_game();
			stop_timer0();
			stop_timer1();
			_delay_ms(500);
			start_timer0();
			start_timer1();
		} else if(serial_input == 'o' || serial_input == 'O') {
			stop_timer0();
			stop_timer1();
			open_game();
			scroll_speed = new_scroll_speed();
			display_ingame_status();
			flash_timer = current_time;
			redraw_background(1);
			flash_powerup();
			_delay_ms(500);
			start_timer0();
			start_timer1();
		} else if(serial_input == 'p' || serial_input == 'P' || paused) {
			stop_timer0();
			stop_timer1();
			serial_input = -1;
			paused = 1;
			while(1) {
				if (serial_input_available()) {
					serial_input = fgetc(stdin);
				}
				if (serial_input == 'p' || serial_input == 'P') {
					paused = 0;
					start_timer1();
					start_timer0();
					break;
				} else if (serial_input == 'n' || serial_input == 'N') {
					paused = 0;
					start_timer0();
					return 1;
				} else if(serial_input == 's' || serial_input == 'S') {
					save_game();
					_delay_ms(500);
				} else if(serial_input == 'o' || serial_input == 'O') {
					open_game();
					scroll_speed = new_scroll_speed();
					display_ingame_status();
					flash_timer = current_time;
					redraw_background(1);
					flash_powerup();
					_delay_ms(500);
					start_timer0();
					start_timer1();
					break;
				}
			}
		} if(serial_input == 'n' || serial_input == 'N') {
			return 1;
		}
		// else - invalid input or we're part way through an escape sequence -
		
		current_time = get_clock_ticks();
		if(!has_car_crashed() && current_time >= last_move_time + scroll_speed) {
			// the allotted time has passed since the last time we scrolled
			// the background, so scroll it now and check whether that means
			// we've finished the lap. (If a crash occurs we will drop out of 
			// the main while loop so we don't need to check for that here.)
			
			scroll_background();
			scrolled = 1;
			//score handling
			if (5 - get_count() <= 0) {
				add_to_score(0);
				} else {
				add_to_score(5 - get_count());
			}
			reset_count();
			update_score();
			// lap handling
			if(has_lap_finished()) {
				uint8_t complete;
				complete = handle_new_lap();	// Pauses until a button is pushed
				if (complete) {
					// complete will be 1 if user pressed n or 2 if player won
					return complete;
				}
				// Reset the time of the last scroll
				last_move_time = get_clock_ticks();
				scroll_speed = new_scroll_speed();
			} else {
				last_move_time = current_time;
			}
		}
	}
	
	// If we get here the car has crashed.
	stop_timer1();
	player_killed();
	_delay_ms(500);
	if (get_lives() > 0) {
		uint8_t complete;
		start_over();
		complete = play_game();
		return complete;
	}
	return 0;
}
//////////////////////////////////////////////////////////////////////////
// IF USER PRESSES 'N' KEY WE SHOULD END UP HERE BY RETURNING 1 FROM
// PLAY_GAME() IF GAME EXITS WITHOUT 'N' WE SKIP THIS FUNCTION
//////////////////////////////////////////////////////////////////////////

void user_interrupt(uint8_t usr) {
	init_score();
	update_score();
}


//////////////////////////////////////////////////////////////////////////////
// HANDLERS FROM GAME COMPLETION, EITHER DEATH OR LAP COMPLETE
//////////////////////////////////////////////////////////////////////////////

uint8_t handle_game_over(uint8_t complete) {
	//compare score contains a write method to eeprom if it's different
	compare_score_to_highscore();
	display_ingame_status();
	set_highscores();
	display_highscore();
	
	if (complete == 2) {
		move_cursor(10, 14);
		printf_P(PSTR("Game Complete"));
		move_cursor(10, 15);
		printf_P(PSTR("Yours is the glory this day"));
	} else {
		move_cursor(10,14);
		printf_P(PSTR("GAME OVER   "));
		move_cursor(10,15);
		printf_P(PSTR("Press a button or 'n' to start again"));
	}
	
	while(button_pushed() == -1) {
		if (serial_input_available()) {
			serial_input = fgetc(stdin);
		}
		if ((serial_input == 'n') || (serial_input == 'N')) {
			serial_input = -1;
			return 0;
		} else if(serial_input == 'o' || serial_input == 'O') {
			stop_timer0();
			open_game();
			display_ingame_status();
			redraw_background(1);
			set_alive();
			flash_powerup();
			_delay_ms(500);
			start_timer0();
			start_timer1();
			return 1;
		}
	}
	return 0;
}

// returns 1 if user pressed n or 2 if game complete
uint8_t handle_new_lap(void) {
	// check for game completion
	if (get_laps() == 9) {
		return 2;
	}
	// stop timer and pdate the game status for completing
	// the lap
	stop_timer1();
	bonus_life();
	inc_laps();
	timer = get_stopwatch();
	if (timer/10 < 120) {
		add_to_score(120 - timer/10);
	}
	
	display_ingame_status();
	
	// print victory message to user
	move_cursor(10,14);
	printf_P(PSTR("LAP COMPLETE"));
	move_cursor(10,15);
	printf_P(PSTR("Now get ready for the next challenge!"));
	
	// handle the led display
	ledmatrix_clear();
	init_scrolling_display();
	switch(random()%4) {
		case 0: set_text_colour(COLOUR_LIGHT_ORANGE); break;
		case 1: set_text_colour(COLOUR_RED); break;
		case 2: set_text_colour(COLOUR_YELLOW); break;
		case 3: set_text_colour(COLOUR_GREEN); break;
	}
	
	sprintf(level, "level %d", get_laps()+1);
	set_scrolling_display_text(level);	
	
	while(scroll_display()) {
		if (serial_input_available()) {
			serial_input = fgetc(stdin);
		}
		if (serial_input == 'n' || serial_input == 'N') {
			return 1;
			} else if(serial_input == 'o' || serial_input == 'O') {
			stop_timer0();
			open_game();
			display_ingame_status();
			redraw_background(1);
			set_alive();
			_delay_ms(500);
			start_timer0();
			start_timer1();
			return 0;
		}
		_delay_ms(100);
	}
	
	//start the next lap
	start_over();
	// Delay for half a second
	reset_count();
	_delay_ms(500);
	init_timer1();
	return 0;
}

void save_game(void) {
	uint8_t bytes[9];
	uint16_t words[2];
	uint32_t dwords[5];
	
	bytes[0] = get_lives();
	bytes[1] = get_laps();
	bytes[2] = get_speed() - bytes[1];
	bytes[3] = paused;
	bytes[4] = get_invincible();
	bytes[5] = get_scroll_position();
	bytes[6] = get_car_column();
	bytes[7] = get_puprow();
	bytes[8] = get_pupcol();
	
	words[0] = get_count();
	words[1] = get_score();
	
	dwords[0] = get_clock_ticks();
	dwords[1] = get_stopwatch();
	dwords[2] = get_invinc_timer();
	dwords[3] = current_time;
	dwords[4] = last_move_time;

	for (uint8_t j = 0; j < 9; j++) {
		uint8_t* i = (uint8_t *) 110 + j;
		eeprom_update_byte(i, bytes[j]);
	}
		
	for (uint8_t j = 0; j < 2; j++) {
		uint16_t* i = (uint16_t *) 130 + (j * 2);
		eeprom_update_word(i, words[j]);
	}
		
	for (uint8_t j = 0; j < 5; j++) {
		uint32_t* i = (uint32_t *) 140 + (j * 4);
		eeprom_update_dword(i, dwords[j]);
	}
	eeprom_update_byte((uint8_t*) 201, 123);
	save_temp_data();
}

void open_game(void) {
	uint8_t game_present_sig = 123;
	uint8_t bytes[9];
	uint16_t words[2];
	uint32_t dwords[5];
	
	uint8_t eepval = eeprom_read_byte((uint8_t *) 201);
	if (eepval != game_present_sig) {
		return;	
	}
	
	for (uint8_t j = 0; j < 9; j++) {
		uint8_t* i = (uint8_t *) 110 + j;
		bytes[j] = eeprom_read_byte(i);
	}
		
	for (uint8_t j = 0; j < 2; j++) {
		uint16_t* i = (uint16_t *) 130 + (j * 2);
		words[j] = eeprom_read_word(i);
	}
		
	for (uint8_t j = 0; j < 5; j++) {
		uint32_t* i = (uint32_t *) 140 + (j * 4);
		dwords[j] = eeprom_read_dword(i);
	}
	
	set_lives(bytes[0]);
	set_laps(bytes[1]);
	set_speed(bytes[2]);
	paused = bytes[3];
	set_invincible(bytes[4]);
	set_scroll_position(bytes[5]);
	set_car_column(bytes[6]);
	set_puprow(bytes[7]);
	set_pupcol(bytes[8]);

	set_count(words[0]);
	set_score(words[1]);
	
	set_clock_ticks(dwords[0]);
	set_stopwatch(dwords[1]);
	set_invinc_timer(dwords[2]);
	current_time = dwords[3];
	last_move_time = dwords[4];
	load_temp_data();
}